/*
 * Test directory reading operations, exploring concurrency issues.
 *
 * Copyright (C) 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



static int count_files(const char *filename, int *found)
{
  file_t *reader = client_opendir(filename);
  struct dirent *dirent;

  if (!client_opened(reader))
  {
    printf("Failed to open %s: %s\n", filename, reader != NULL ? l4sys_errtostr(reader->error) : "(no reader)");
    client_close(reader);
    return 1;
  }

  *found = 0;

  while ((dirent = client_readdir(reader)) != NULL)
  {
    free(dirent);
    (*found)++;
  }

  client_close(reader);
  return 0;
}

int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("Need a directory name.\n");
    return 1;
  }

  char *filename = argv[1];
  int original;

  if (count_files(filename, &original))
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  printf("Files found: %d\n", original);

  /* Operate on files with sequence numbering in a certain range. */

  int filenum_base = 1000, filenum = filenum_base;
  int to_create = 100, remaining = to_create;
  char buffer[strlen(filename) + strlen("/file-XXXX.txt") + 10];
  char data[256];
  int found = 0;
  file_t *files[to_create];

  /* Open a directory for reading. */

  printf("Opening %s...\n", filename);

  file_t *reader = client_opendir(filename);

  if (!client_opened(reader))
  {
    client_close(reader);
    printf("Could not read from directory.\n");
    return 1;
  }

  /* Read entries from the directory. */

  struct dirent *dirent;

  printf("Reading...\n");

  while ((dirent = client_readdir(reader)) != NULL)
  {
    free(dirent);
    found++;

    /* Create files while getting the listing. */

    if (remaining)
    {
      sprintf(buffer, "%s/file-%d.txt", filename, filenum);

      file_t *file = client_open(buffer, O_RDWR | O_CREAT);

      if (!client_opened(file))
      {
        client_close(file);
        printf("Could not open file: %s\n", buffer);
        return 1;
      }

      /* Write something to each file. */

      sprintf(data, "Data in file-%d.txt", filenum);
      client_write(file, data, strlen(data));

      /* Remember the file for later. */

      files[filenum - filenum_base] = file;

      remaining--;
      filenum++;
    }
  }

  printf("Files found: %d\n", found);

  /* Re-read, counting files. */

  if (count_files(filename, &found))
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  printf("Files found: %d\n", found);

  /* Remove some files. */

  for (filenum = 1000, remaining = to_create; remaining > 0; remaining--, filenum++)
  {
    sprintf(buffer, "%s/file-%d.txt", filename, filenum);

    long err = client_remove(buffer);

    if (err)
    {
      printf("Could not remove file: %s\n", buffer);
      return 1;
    }
  }

  /* Re-read, counting files. */

  if (count_files(filename, &found))
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  printf("Files found: %d\n", found);

  /* Read from the still-open but now removed files. */

  found = 0;

  for (filenum = 1000, remaining = 0; remaining < to_create; remaining++, filenum++)
  {
    file_t *file = files[filenum - filenum_base];
    char read_data[256];

    client_seek(file, SEEK_SET, 0);
    offset_t nread = client_read(file, read_data, 256);

    if (nread)
    {
      read_data[nread] = '\0';
      sprintf(data, "Data in file-%d.txt", filenum);

      if (!strcmp(data, read_data))
        found++;
    }
  }

  printf("Files still open: %d (%d)\n", found, to_create);
  printf("All created files still open: %s\n", found == to_create ? "True" : "False");

  /* Close the files. */

  for (remaining = 0; remaining < to_create; remaining++)
    client_close(files[remaining]);

  printf("Closed files.\n");

  /* Re-read, counting files. */

  if (count_files(filename, &found))
  {
    printf("Could not read from directory.\n");
    return 1;
  }

  printf("Files found: %d (%d)\n", found, original);
  printf("Removed all surplus files: %s\n", found == original ? "True" : "False");
  printf("End of test.\n");
  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
